/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "qp_port.h"

#include <stdint.h>

Q_DEFINE_THIS_MODULE(hash)

#define Hash_Get16bits(data_)   ((((uint32_t)(((const uint8_t *)(data_))[1])) << 8)\
                                   +(uint32_t)(((const uint8_t *)(data_))[0]) )

/*..........................................................................*/
uint32_t Hash_Create(const uint8_t * data, uint32_t length, uint32_t hash)
{
	uint32_t tmp, rem;

	Q_ASSERT(data != 0);
	
    rem = length & 3;
    length >>= 2;

    /* main loop */
    for (;length > 0; length--) {
        hash  += Hash_Get16bits(data);
        tmp    = (Hash_Get16bits(data+2) << 11) ^ hash;
        hash   = (hash << 16) ^ tmp;
        data  += 2*sizeof (uint16_t);
        hash  += hash >> 11;
    }

    /* handle end cases */
    switch (rem) {
    case 3: {
        hash += Hash_Get16bits(data);
        hash ^= hash << 16;
        hash ^= data[sizeof (uint16_t)] << 18;
        hash += hash >> 11;
        break;
    }
    case 2: {
        hash += Hash_Get16bits(data);
        hash ^= hash << 11;
        hash += hash >> 17;
        break;
    }
    case 1: {
        hash += *data;
        hash ^= hash << 10;
        hash += hash >> 1;
    }
    }

    /* force "avalanching" of final 127 bits */
    hash ^= hash << 3;
    hash += hash >> 5;
    hash ^= hash << 4;
    hash += hash >> 17;
    hash ^= hash << 25;
    hash += hash >> 6;

    return hash;
}
